package PureMessage::Policy::CreditCard;

use strict;
use warnings;

use base 'PureMessage::Policy::Plugin';
use PureMessage::Policy::Words qw(message_words);
use PureMessage::Policy::CreditCard::Config qw(%config);
use PureMessage::Policy::Plugin::Utils qw(
        expand_localized_tmpl
        replace_nested_part
        set_attachments 
        );
use PureMessage::Policy::MessageText qw(match_attachment_text
                                        per_attachment_byte_limit_exceeded
                                        per_message_time_limit_exceeded
                                        current_attachment_part_id
        );

our @EXPORT_OK = qw(pmx_credit_card);
our $VERSION   = '5.500';

my %commands = (
    pmx_credit_card                   => \&pmx_credit_card,
);

sub new {
    my $class = shift;
    my %opts = @_;
    return bless \%opts, $class;
}

sub commands {
    my $self = shift;
    return %commands;
}

sub pmx_credit_card {
    my ($msg,$args) = @_;

    my $no_early_out = $args->tag('scanall') ? 1 : 0;

    # Clear the cantscan hash attribute
    $msg->set_attr('cantscan_map', [{'type' => 'cc', 'ids' => []}], 1);
    my $cantscan_map = { type => 'cc', ids => [] };

    eval {
        message_words(sub {
            my (undef, $part_id, $words) = @_;
            my $str = join('', @$words);
            my $res = _credit_card_check($str);
            if ($res == 1) {
                die "--matched--$part_id--";
            } 
            elsif($res == -1) {
                die "--limit--$part_id--";
            }
        }, $msg, overlap => 15, word_re => qr/[-_\s]+/);
    };

    if ($@) {
        if ($@ =~ /--matched--/) {
            $msg->log("DEBUG", "pmx_credit_card: credit card found");
            return 1;
        } 
        elsif ($@ =~ /--limit--([^-]+)--/) {
            my $part_id = $1;

            $msg->set_attr('cc_limit_exceeded', 1);
            $msg->log("DEBUG", "pmx_credit_card: credit card check limit exceeded");
            
            push(@{ $cantscan_map->{ids} }, {'sav_id' => 'SOPHOS_CC_NUMBER_LIMIT', 'part_id' => "$part_id"} );
            $msg->set_attr('cantscan_map', [$cantscan_map], 1);

            return 1;
        }
        $msg->log("DEBUG", "pmx_credit_card: Unknown error");
        die $@;     # a real error was caught, not our die
    }
    

    if ($args->tag('attachments')) {
        my @matching_parts;
        my @non_matching_parts;

        my @matches = match_attachment_text($msg,
                            sub {
                                my $attachment_text = shift;

                                # Remove any spaces, dashes or underscores from the text
                                $attachment_text =~ s/[\-_\s]+//g;

                                my $result = _credit_card_check($attachment_text);

                                if ($result == -1) {
                                    push(@{ $cantscan_map->{ids} }, {
                                            'sav_id' => 'SOPHOS_CC_NUMBER_LIMIT', 
                                            'part_id' => current_attachment_part_id(),
                                        } );
                                    $msg->set_attr('cc_limit_exceeded', 1);
                                    $msg->log("DEBUG", "pmx_credit_card :attachments - credit card check limit exceeded");
                                }

                                # Returns an array to make match_attachment_text() happy.
                                # (match_attachment_text() is expecting an array of length 2 to be returned
                                # from the matcher callback) 
                                # We never want to return an actual credit card number for obvious privacy
                                # concerns.
                                return ($result, '');
                            },
                matching_parts     => \@matching_parts,
                non_matching_parts => \@non_matching_parts,
                no_early_out       => $no_early_out,
            );

        my $modified_cantscan = $msg->get_attr('cantscan_map');
        if (defined $modified_cantscan and my $last_cantscan = $modified_cantscan->[-1]) {
            push @{$last_cantscan->{ids}}, @{$cantscan_map->{ids}};
            $msg->set_attr('cantscan_map', $modified_cantscan, 1);

        } else {
            $msg->set_attr('cantscan_map', [$cantscan_map], 1);
        }

        if (@matches) {
            $msg->log("DEBUG", "pmx_credit_card :attachments - credit card found");
            $msg->template("MATCHED_FILE", $matches[0]->{file});
            set_attachments($msg, \@matching_parts, \@non_matching_parts);

            return 1;
        }

        set_attachments($msg, \@matching_parts, \@non_matching_parts);

        return (pmx_credit_card_limit_exceeded($msg)
            || $msg->{per_message_savi_cantscan_errors});
    }

    return 0;
}

sub pmx_credit_card_limit_exceeded {
    my $msg = shift;
    return $msg->get_attr('cc_limit_exceeded')
        || per_attachment_byte_limit_exceeded($msg)
        || per_message_time_limit_exceeded($msg);
}

sub _credit_card_check {
	my $str = shift;

    # Maximum number of cc checks in a single message
    my $cc_check_limit = $config{creditcard_number_limit};

    my $number_limit = $cc_check_limit * 2;  # Since we step through our matches 2 at a time below
    my @potential_matches =  ($str =~ m/([3546](?=(\d{12,15})))/g);

    for (my $idx = 0; $idx < @potential_matches - 1; $idx+= 2) {
        my $first_digit   = $potential_matches[$idx];
        my $potential_num = $first_digit . $potential_matches[$idx + 1];

        if ($idx >= $number_limit) {
            return -1;
        }

        # We must check each CC string multiple times incase there are nested CC numbers
        my @cc_patterns = (
	        '(5[1-5][\d]{14})',          # Mastercard / Diner's Club US & Canada
	        '(4[\d]{12}([\d]{3})?)',     # Visa
	        '(3[467][\d]{13})',          # AmEx
	        '(6011[\d]{12})',            # Discover
	        '(65[\d]{14})',              # Discover
        );
        foreach my $cc_check (@cc_patterns) {
            next unless($potential_num =~ qr/$cc_check/);

	        my $cc_num = $1;

	        # Do a Luhn check on the number to make sure it's legit
	        # (based on the code from Busines::CreditCard)
	        my $weight;
	        my $sum = 0;
	        for (my $i = 0; $i < length($cc_num) - 1; $i++) {
	            $weight = substr($cc_num, -1 * ($i + 2), 1) * (2 - ($i % 2));
	            $sum += (($weight < 10) ? $weight : ($weight - 9));
	        }

	        if (substr($cc_num, -1) == (10 - $sum % 10) % 10) {
	            return 1;
	       	}
	    }
	}	
	return 0;
}

sub expand_templates {
    my ($msg, $tmpl, $part_id) = @_;

    use PureMessage::Template qw(expand_template);
    $tmpl = expand_template(
        $tmpl,
        vars => {
                header_subject	=> scalar $msg->header('subject'),
                qid             => scalar $msg->minfo_value('Queue-Id'),
        },
    );
    return $tmpl;
}

1;

